#!/usr/bin/python3
"""
Master Node that manages connections with the cu nodes.
"""
#######################        MANDATORY IMPORTS         #######################

#######################         GENERIC IMPORTS          #######################
import threading
from typing import List

#######################       THIRD PARTY IMPORTS        #######################

#######################    SYSTEM ABSTRACTION IMPORTS    #######################
from rfb_logger_tool import sys_log_logger_get_module_logger, SysLogLoggerC, Logger

#######################       LOGGER CONFIGURATION       #######################
if __name__ == '__main__':
    cycler_logger = SysLogLoggerC(file_log_levels='./log_config.yaml')
log: Logger = sys_log_logger_get_module_logger(__name__)

#######################          PROJECT IMPORTS         #######################
from rfb_shared_tool import SysShdIpcChanC, SysShdNodeC, SysShdNodeStatusE
from rfb_cycler_datatypes.comm_data import (CommDataCuC,CommDataRegisterTypeE,
                                                        CommDataHeartbeatC, CommDataDeviceC,
                                                        CommDataMnCmdDataC, CommDataMnCmdTypeE)

#######################          MODULE IMPORTS          #######################
from .mn_broker_client import BrokerClientC
from .mn_db_facade import DbFacadeC

#######################              ENUMS               #######################

######################             CONSTANTS              ######################
from .context import (DEFAULT_MN_NODE_NAME, DEFAULT_NODE_PERIOD)
MN_REQS_CHAN_NAME = 'mn_reqs'
MN_DATA_CHAN_NAME = 'mn_data'

#######################             CLASSES              #######################

class MnManagerNodeC(SysShdNodeC): # pylint: disable=abstract-method
    '''
    Cu Manager Class to instanciate a CU Manager Node
    '''
    def __init__(self, working_flag : threading.Event,
        cycle_period : int= DEFAULT_NODE_PERIOD) -> None:
        '''
        Initialize the node.
        '''
        super().__init__(name=DEFAULT_MN_NODE_NAME , cycle_period=cycle_period,
                working_flag=working_flag)

        self.mn_req_chan : SysShdIpcChanC = SysShdIpcChanC(name=MN_REQS_CHAN_NAME,
                                                            max_message_size=350,
                                                            max_msg=350)
        self.mn_data_chan : SysShdIpcChanC = SysShdIpcChanC(name=MN_DATA_CHAN_NAME,
                                                            max_message_size=350,
                                                            max_msg=350)
        self.db_facha : DbFacadeC = DbFacadeC()
        avail_cus : List[int] = self.db_facha.get_available_cus()
        log.info(f"Already connected devices: {avail_cus}")

        self.client_mqtt : BrokerClientC = BrokerClientC(error_callback=self.error_cb,
                                                         register_cb=self.register_cb,
                                                         heartbeat_cb=self.heartbeat_cb,
                                                         inform_dev_cb=self.detect_devices_cb,
                                                         avail_cus=avail_cus)

    def error_cb(self, data) -> None:
        '''Error callback for Broker.

        Args:
            data ([type]): [description]
        '''
        log.critical(f'Error in Broker Client: {data}')
        self.status = SysShdNodeStatusE.COMM_ERROR


    def register_cb(self, cu_info : CommDataCuC) -> None:
        '''This callback is called when the CU is received.

        Args:
            cu_info (CommDataCuC): [description]
        '''
        if cu_info.msg_type is CommDataRegisterTypeE.DISCOVER:
            found_cu = self.db_facha.get_cu_by_mac(cu_info.mac)
            if found_cu is not None:
                cu_info.cu_id = found_cu
                cu_info.msg_type = CommDataRegisterTypeE.OFFER
                log.info(f"{cu_info.msg_type.name}s cu_id {cu_info.cu_id} for device already "
                        + "found in database with MAC:"
                        + f"{cu_info.mac}")
            else:
                new_cu_id = self.db_facha.get_last_cu_id()
                cu_info.cu_id = new_cu_id + 1
                cu_info.msg_type = CommDataRegisterTypeE.OFFER
                log.info(f"{cu_info.msg_type.name}s cu_id {cu_info.cu_id} for device with MAC:"
                        + f"{cu_info.mac}")
            self.client_mqtt.publish_inform(cu_info)
        elif cu_info.msg_type is CommDataRegisterTypeE.REQUEST:
            if self.db_facha.is_cu_registered(cu_info):
                cu_info.msg_type = CommDataRegisterTypeE.ACK
                log.info(f"Send {cu_info.msg_type.name}. Already registered CU: {cu_info.cu_id} "
                            + f"with MAC: {cu_info.mac}")
                self.client_mqtt.publish_inform(cu_info)
            else:
                self.db_facha.register_cu(cu_info)
                try:
                    self.db_facha.commit()
                except Exception as err:
                    log.error(f"Error on commiting new CU: {err}")
                else:
                    cu_info.msg_type = CommDataRegisterTypeE.ACK
                    log.info(f"Send {cu_info.msg_type.name}. Registered new CU: {cu_info.cu_id} "
                            + f"with MAC: {cu_info.mac}")
                    self.client_mqtt.publish_inform(cu_info)
        else:
            log.debug(f"Inconsistent register message: {cu_info.msg_type}. Ignore it.")


    def heartbeat_cb(self, heartbeat : CommDataHeartbeatC) -> None:
        '''Handle heartbeat data.

        Args:
            hb (CommDataHeartbeatC): [description]
        '''
        self.db_facha.update_heartbeat(heartbeat=heartbeat)


    def detect_devices_cb(self, cu_id : int, devices : List[CommDataDeviceC]) -> None:
        '''Callback called when a device is detected .

        Args:
            cu_id (int): [description]
            devices (List[CommDataDeviceC]): [description]
        '''
        log.info(f"Devices detected for [{cu_id}]: {devices}")
        self.db_facha.update_devices(cu_id, devices)
        msg_data = CommDataMnCmdDataC(cmd_type=CommDataMnCmdTypeE.INF_DEV,
                                            cu_id=cu_id, devices=devices)
        self.mn_data_chan.send_data(msg_data)


    def apply_cmds(self) -> None:
        '''Applies all registered commands to the broker.
        '''
        cmd = self.mn_req_chan.receive_data_unblocking()
        if isinstance(cmd, CommDataMnCmdDataC):
            log.warning(f"Applying {cmd.cmd_type.name} cmd")
            if cmd.cmd_type is CommDataMnCmdTypeE.LAUNCH:
                self.client_mqtt.publish_launch(cu_id=cmd.cu_id, cs_id=cmd.cs_id)
            elif cmd.cmd_type is CommDataMnCmdTypeE.REQ_DETECT:
                self.client_mqtt.publish_req_devices(cu_id=cmd.cu_id)


    def process_iteration(self) -> None:
        '''Perform a single iteration of the protocol.
        '''
        log.debug("New Processing iteration")
        self.apply_cmds()
        self.db_facha.track_avail_cu()
        self.db_facha.commit()
        self.client_mqtt.process_incomming_msg()


    def stop(self) -> None:
        '''Stop the stream .
        '''
        self.client_mqtt.close()

#######################            FUNCTIONS             #######################



#!/usr/bin/python3
"""
Wrapper for the MQTT client
"""
#######################        MANDATORY IMPORTS         #######################

#######################         GENERIC IMPORTS          #######################
from typing import Callable, List
from pickle import dumps, loads

#######################       THIRD PARTY IMPORTS        #######################

#######################    SYSTEM ABSTRACTION IMPORTS    #######################
from rfb_logger_tool import sys_log_logger_get_module_logger, SysLogLoggerC, Logger

#######################       LOGGER CONFIGURATION       #######################
if __name__ == '__main__':
    cycler_logger = SysLogLoggerC(file_log_levels='../log_config.yaml')
log: Logger = sys_log_logger_get_module_logger(__name__)

#######################          PROJECT IMPORTS         #######################
from rfb_cycler_datatypes.comm_data import CommDataCuC, CommDataRegisterTypeE,\
    CommDataHeartbeatC, CommDataDeviceC
from rfb_driver_mqtt import DrvMqttDriverC

#######################          MODULE IMPORTS          #######################

#######################              ENUMS               #######################

######################             CONSTANTS              ######################
from .context import (DEFAULT_CRED_FILEPATH)

_REGISTER_TOPIC = '/register'
_INFORM_TOPIC = '/inform_reg'
_SUFFIX_RX_DET_DEV = '/detected_dev'
_SUFFIX_RX_HB = '/heartbeat'
_SUFFIX_TX_DET = '/req_detect'
_SUFFIX_TX_LAUNCH = '/launch'

#######################             CLASSES              #######################

class BrokerClientC():
    """
    Broker Client Class to instanciate a Broker Client object
    """
    def __init__(self, error_callback : Callable,
                 register_cb : Callable, heartbeat_cb : Callable,
                 inform_dev_cb : Callable, avail_cus : List[int]) -> None:

        self.mqtt : DrvMqttDriverC = DrvMqttDriverC(error_callback=error_callback,
                                                    cred_path=DEFAULT_CRED_FILEPATH)

        self.__register_cb : Callable = register_cb
        self.__heartbeat_cb : Callable = heartbeat_cb
        self.__inform_dev_cb : Callable = inform_dev_cb
        self.mqtt.subscribe(topic=_REGISTER_TOPIC, callback=self.process_register)
        self.mqtt.subscribe(topic=_INFORM_TOPIC, callback=self.process_register)
        for cu_id in avail_cus:
            self.__subscribe_cu(cu_id=cu_id)


    def __subscribe_cu(self, cu_id : int) -> None:
        hb_topic: str = f'/{cu_id}{_SUFFIX_RX_HB}'
        self.mqtt.subscribe(topic=hb_topic, callback=self.process_heartbeat)
        det_dev_topic: str = f'/{cu_id}{_SUFFIX_RX_DET_DEV}'
        self.mqtt.subscribe(topic=det_dev_topic, callback=self.process_det_dev)


    def process_register(self, raw_data : bytearray) -> None:
        '''
        Process the received data from the broker

        Args:
            raw_data (bytearray): Received data
        '''
        cu_info = loads(raw_data)
        if isinstance(cu_info, CommDataCuC):
            log.info(f'Received register cmd: {cu_info.msg_type}')
            self.__register_cb(cu_info)
        else:
            log.error(f'Invalid data received: {type(cu_info)}: {cu_info}')


    def process_heartbeat(self, raw_data : bytearray) -> None:
        '''Process a heartbeat from the device.

        Args:
            raw_data (bytearray): [description]
        '''
        heartbeat : CommDataHeartbeatC = loads(raw_data)
        log.debug(f'Heartbeat received from: {heartbeat.cu_id}')
        self.__heartbeat_cb(heartbeat)


    def process_det_dev(self, raw_data : bytearray) -> None:
        '''Process a device from a CUDA device.

        Args:
            raw_data (bytearray): [description]
        '''
        devices : List[CommDataDeviceC] = loads(raw_data)
        if len(devices) > 0:
            cu_id = devices[0].cu_id
            log.info(f"Received detected from [{cu_id}]")
            self.__inform_dev_cb(cu_id, devices)
        else:
            log.warning(f"No devices detected from any CU")

    def publish_inform(self, cu_info : CommDataCuC) -> None:
        '''Publish the inform data.

        Args:
            cu_info (CommDataCuC): [description]
        '''
        if cu_info.msg_type is CommDataRegisterTypeE.ACK:
            self.__subscribe_cu(cu_id=cu_info.cu_id)

        raw_cu: bytes = dumps(cu_info)
        self.mqtt.publish(topic=_INFORM_TOPIC, data=raw_cu)


    def publish_launch(self, cu_id : int, cs_id : int) -> None:
        '''Publish a launch to the device.

        Args:
            cu_id (int): [description]
            cs_id (int): [description]
        '''
        launch_topic: str = f'/{cu_id}{_SUFFIX_TX_LAUNCH}'
        self.mqtt.publish(topic=launch_topic, data=cs_id)


    def publish_req_devices(self, cu_id : int) -> None:
        '''Publish the request devices for the given cu_id.

        Args:
            cu_id (int): [description]
        '''
        req_topic: str = f'/{cu_id}{_SUFFIX_TX_DET}'
        self.mqtt.publish(topic=req_topic, data=cu_id)


    def process_incomming_msg(self) -> None:
        '''Process possible incoming messages.
        '''
        self.mqtt.process_data()


    def close(self) -> None:
        '''Close the broker client.
        '''
        self.mqtt.close()


#######################            FUNCTIONS             #######################



#!/usr/bin/python3
"""
Wrapper for the MQTT client
"""
#######################        MANDATORY IMPORTS         #######################
from __future__ import annotations
#######################         GENERIC IMPORTS          #######################
from typing import List
from datetime import datetime, timedelta

#######################       THIRD PARTY IMPORTS        #######################
from sqlalchemy.sql.expression import update, select

#######################    SYSTEM ABSTRACTION IMPORTS    #######################
from rfb_logger_tool import sys_log_logger_get_module_logger, SysLogLoggerC, Logger

#######################       LOGGER CONFIGURATION       #######################
if __name__ == '__main__':
    cycler_logger = SysLogLoggerC(file_log_levels='../log_config.yaml')
log: Logger = sys_log_logger_get_module_logger(__name__)

#######################          MODULE IMPORTS          #######################

#######################          PROJECT IMPORTS         #######################
from rfb_cycler_datatypes.comm_data import (CommDataCuC, CommDataHeartbeatC, CommDataDeviceC)
from rfb_driver_db import (DrvDbDetectedDeviceC, DrvDbSqlEngineC, DrvDbTypeE,
                            DrvDbComputationalUnitC, DrvDbAvailableCuE, DrvDbConnStatusE)

#######################              ENUMS               #######################
######################             CONSTANTS              ######################
from .context import (DEFAULT_TIMEOUT_BETWEEN_CONNECTIONS)

#######################             CLASSES              #######################

class DbFacadeC:
    '''This function is used to create a DbFacade class for the Django backend .
    '''

    def __init__(self) -> None:
        self.database : DrvDbSqlEngineC  = DrvDbSqlEngineC(db_type=DrvDbTypeE.MASTER_DB,
                                            config_file='./devops/mn_manager/.cred.yaml')
        self.last_cu_id = -1


    def get_last_cu_id(self) -> int:
        '''Get the CUDA CU ID for this simulation.

        Returns:
            int: [description]
        '''
        stmt = select(DrvDbComputationalUnitC.CUID)\
                        .order_by(DrvDbComputationalUnitC.CUID.desc())\
                        .limit(1)
        res = self.database.session.execute(stmt).first()
        if res is not None:
            self.last_cu_id = res[0]

        return self.last_cu_id

    def get_available_cus(self) -> List[int]:
        '''Returns a list of available CUDA units.

        Returns:
            List[int]: [description]
        '''
        stmt = select(DrvDbComputationalUnitC.CUID)\
                        .filter(DrvDbComputationalUnitC.Available == DrvDbAvailableCuE.ON.value)\
                        .order_by(DrvDbComputationalUnitC.CUID.asc())
        res = self.database.session.execute(stmt).fetchall()
        cus = []
        for c_unit in res:
            cus.append(c_unit[0])
        return cus


    def is_cu_registered(self, cu_info : CommDataCuC) -> bool:
        '''Check if a CU is already registered.

        Args:
            cu_info (CommDataCuC): [description]

        Returns:
            bool: [description]
        '''
        result = False
        stmt = select(DrvDbComputationalUnitC)\
                        .where(DrvDbComputationalUnitC.MAC == cu_info.mac)
        db_result = self.database.session.execute(stmt).first()
        if db_result is not None:
            result = True
        return result


    def get_cu_by_mac(self, mac : int) -> int|None:
        '''Returns the CU_ID from database for the given mac address.
        Returns None if not found.

        Args:
            mac (int): [description]

        Returns:
            result (int|None): CU_ID retrieved from database. None if not found
        '''
        result : int = None
        stmt = select(DrvDbComputationalUnitC)\
                        .where(DrvDbComputationalUnitC.MAC == mac)
        db_result = self.database.session.execute(stmt).first()
        if db_result is not None:
            query_result : DrvDbComputationalUnitC = db_result[0]
            result = query_result.CUID
        return result


    def register_cu(self, cu_info : CommDataCuC) -> None:
        '''Register a CU data unit.

        Args:
            cu_info (CommDataCuC): [description]
        '''
        log.info(f"Registering new CU: {cu_info}")
        self.last_cu_id += 1
        cu_db = DrvDbComputationalUnitC()
        cu_db.CUID = self.last_cu_id
        cu_db.MAC = cu_info.mac
        cu_db.HostName = cu_info.hostname
        cu_db.User = cu_info.user
        cu_db.IP = cu_info.ip
        cu_db.Port = cu_info.port
        cu_db.LastConnection = datetime.utcnow()
        cu_db.Available = DrvDbAvailableCuE.ON.value
        self.database.session.add(cu_db)


    def update_heartbeat(self, heartbeat : CommDataHeartbeatC) -> None:
        '''Update the commData of the commdata.

        Args:
            hb (CommDataHeartbeatC): [description]
        '''
        stmt = update(DrvDbComputationalUnitC)\
                        .where(DrvDbComputationalUnitC.CUID == heartbeat.cu_id)\
                        .values(LastConnection= datetime.utcnow())
        self.database.session.execute(stmt)


    def update_devices(self, cu_id : int, devices : List[CommDataDeviceC]) -> None:
        '''Add or Update the list of devices in the database.

        Args:
            cu_id (int): [description]
            devices (List[CommDataDeviceC]): [description]
        '''
        select_stmt = select(DrvDbDetectedDeviceC)\
                            .where(DrvDbDetectedDeviceC.CUID == cu_id)
        res = self.database.session.execute(select_stmt).fetchall()
        for db_dev in res:
            db_dev : DrvDbDetectedDeviceC = db_dev[0]
            db_dev.ConnStatus = DrvDbConnStatusE.DISCONNECTED.value
            log.debug(f"Set device: {db_dev.__dict__} as disconnected")
            update_stmt = update(DrvDbDetectedDeviceC)\
                            .where(DrvDbDetectedDeviceC.CUID == db_dev.CUID)\
                            .where(DrvDbDetectedDeviceC.CompDevID == db_dev.CompDevID)\
                            .where(DrvDbDetectedDeviceC.SN == db_dev.SN)\
                            .where(DrvDbDetectedDeviceC.LinkName == db_dev.LinkName)\
                            .values(ConnStatus=DrvDbConnStatusE.DISCONNECTED.value)
            self.database.session.execute(update_stmt)
        msg = f"Setting as disconnected all devices in: {cu_id} to update only the connected ones"
        log.info(msg)
        self.commit()
        for device in devices:
            select_stmt = select(DrvDbDetectedDeviceC)\
                            .where(DrvDbDetectedDeviceC.CUID == cu_id)\
                            .where(DrvDbDetectedDeviceC.CompDevID == device.comp_dev_id)\
                            .where(DrvDbDetectedDeviceC.SN == device.serial_number)\
                            .where(DrvDbDetectedDeviceC.LinkName == device.link_name)
            res = self.database.session.execute(select_stmt).first()
            if res is not None:
                db_dev : DrvDbDetectedDeviceC = res[0]
                db_dev.ConnStatus = DrvDbConnStatusE.CONNECTED.value
                log.info(f"Updating device: {device}")
                update_stmt = update(DrvDbDetectedDeviceC)\
                            .where(DrvDbDetectedDeviceC.CUID == cu_id)\
                            .where(DrvDbDetectedDeviceC.CompDevID == device.comp_dev_id)\
                            .where(DrvDbDetectedDeviceC.SN == device.serial_number)\
                            .where(DrvDbDetectedDeviceC.LinkName == device.link_name)\
                            .values(ConnStatus=DrvDbConnStatusE.CONNECTED.value)
                self.database.session.execute(update_stmt)
            else:
                db_dev : DrvDbDetectedDeviceC = DrvDbDetectedDeviceC()
                db_dev.CUID = cu_id
                db_dev.CompDevID = device.comp_dev_id
                db_dev.SN = device.serial_number
                db_dev.LinkName = device.link_name
                db_dev.ConnStatus = DrvDbConnStatusE.CONNECTED.value
                log.info(f"Adding device: {device}")
                self.database.session.add(db_dev)
        log.info(f"Commiting add and/or update all devices in: {cu_id}")
        self.commit()


    def track_avail_cu(self) -> None:
        '''
        Track available CUs.
        '''
        # Update DrvDbComputationalUnitC.Available if elapsed time from LastConnection
        # is less than the defined constant
        last_ts_allowed = datetime.utcnow() - timedelta(seconds=DEFAULT_TIMEOUT_BETWEEN_CONNECTIONS)

        stmt_put_off = update(DrvDbComputationalUnitC)\
            .where(DrvDbComputationalUnitC.Available==DrvDbAvailableCuE.ON.value,\
                DrvDbComputationalUnitC.LastConnection < last_ts_allowed)\
            .values(Available=DrvDbAvailableCuE.OFF.value)

        stmt_put_on = update(DrvDbComputationalUnitC)\
            .where(DrvDbComputationalUnitC.Available==DrvDbAvailableCuE.OFF.value,\
                DrvDbComputationalUnitC.LastConnection > last_ts_allowed)\
            .values(Available=DrvDbAvailableCuE.ON.value)

        self.database.session.execute(stmt_put_off)
        self.database.session.execute(stmt_put_on)

    def commit(self) -> None:
        '''
        Commit changes to the database.

        Raise:
            Exception: if there is an error during the commit.
        '''
        self.database.commit_changes(raise_exception=True)

#######################            FUNCTIONS             #######################



'''
This package contains the classes and functions needed to communicate database and
web server with CUs.
'''

from .mn_manager_node import MN_DATA_CHAN_NAME, MN_REQS_CHAN_NAME, MnManagerNodeC

__all__ = ['MN_DATA_CHAN_NAME', 'MN_REQS_CHAN_NAME', 'MnManagerNodeC']



#!/usr/bin/python3
'''
This module manages the constants variables.
Those variables are used in the scripts inside the module and can be modified
in a config yaml file specified in the environment variable with name declared
in system_config_tool.
'''

#######################        MANDATORY IMPORTS         #######################
from __future__ import annotations
#######################         GENERIC IMPORTS          #######################

#######################      SYSTEM ABSTRACTION IMPORTS  #######################
from rfb_logger_tool import Logger, sys_log_logger_get_module_logger
log: Logger = sys_log_logger_get_module_logger(__name__)

#######################       THIRD PARTY IMPORTS        #######################

#######################          PROJECT IMPORTS         #######################
from rfb_config_tool import sys_conf_update_config_params

#######################          MODULE IMPORTS          #######################

######################             CONSTANTS              ######################
# For further information check out README.md
DEFAULT_CRED_FILEPATH : str = 'devops/mn_manager/.cred.yaml' #Max number of allowed message per chan
DEFAULT_MN_NODE_NAME: str = 'cu_manager_node'
DEFAULT_NODE_PERIOD: int = 750 # ms # Period of the node
DEFAULT_TIMEOUT_BETWEEN_CONNECTIONS: int =15


CONSTANTS_NAMES = ('DEFAULT_CRED_FILEPATH','DEFAULT_MN_NODE_NAME', 'DEFAULT_NODE_PERIOD',
                    'DEFAULT_TIMEOUT_BETWEEN_CONNECTIONS')
sys_conf_update_config_params(context=globals(),
                              constants_names=CONSTANTS_NAMES)
